import socket, pickle
import numpy as np
from ..agents import PokerAgent, RandomAgent

class PokerGameClient:
	"""  """

	def __init__(self, agent: PokerAgent=RandomAgent()):
		"""  """

		assert agent is not None, 'Invalid agent'
		self.agent = agent

		self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.player = -1
		self.state = None

	def connect(self, host: str='localhost', port: int=25560):
		"""  """

		# Wait for ready state
		self.sock.connect((host, int(port)))
		ready_state = self.sock.recv(1024)
		ready_state: dict= pickle.loads(ready_state)

		assert ready_state.get('status') == 'ready', 'Invalid status'
		self.player = ready_state.get('player', 0)
		print('Connnected to %s:%d as player %d' % (host, port, self.player))

		while True:
			# Receive state from server
			state = self.sock.recv(4096)
			state: dict = pickle.loads(state)
			assert state.get('player') == self.player, 'Received state with wrong player id'

			status = state.get('status', 'playing')
			if status == 'over': break
			elif status == 'playing':
				# Send random action
				active_state = state.get('active_state')
				action = self.agent(active_state)
				action = pickle.dumps((self.player, action))
				self.sock.sendall(action)
			else: raise AssertionError('Invalid status')


class HandRanking:
	""" Enum of hand rankings in poker
	
	A list with a description of
	each type of hand can be found
	here: https://en.wikipedia.org/wiki/List_of_poker_hands#Hand-ranking_categories
	"""

	STRAIGHT_FLUSH = 1
	POKER = 2
	FULL = 3
	FLUSH = 4
	STRAIGHT = 5
	TRIS = 6
	TWO_PAIR = 7
	PAIR = 8
	HIGH = 9
	NONE = 10

	# Strings for each hand type
	as_string = [
		'Five of a Kind',
		'Straight Flush',
		'Four of a Kind',
		'Full House',
		'Flush',
		'Straight',
		'Three of a Kind',
		'Two Pair',
		'One Pair',
		'High Card',
		'Nothing'
	]

class CardSuit:
	""" Enum with the suits of playing cards """

	SPADES = 0x0
	HEARTS = 0x1
	DIAMONDS = 0x2
	CLUBS = 0x3

	NUM_SUITS = 4
	MAX_SUIT = 3

	as_string = ['Spades', 'Hearts', 'Diamonds', 'Clubs']
	as_symbol = ['\u2660', '\u2665', '\u2666' '\u2663']

	from_symbol = {
		'S': SPADES,
		'H': HEARTS,
		'D': DIAMONDS,
		'C': CLUBS,
	}

class CardRank:
	""" Enum with ranks of playing cards """

	ONE = 0x0
	TWO = 0x1
	THREE = 0x2
	FOUR = 0x3
	FIVE = 0x4
	SIX = 0x5
	SEVEN = 0x6
	EIGHT = 0x7
	NINE = 0x8
	TEN = 0x9
	JACK = 0xa
	QUEEN = 0xb
	KING = 0xc
	ACE = 0xd

	NUM_RANKS = 13
	MAX_RANK = 13

	as_string = ['Ace', 'Two', 'Three', 'Four', 'Five', 'Six', 'Seven', 'Eight', 'Nine', 'Ten', 'Jack', 'Queen', 'King', 'Ace']
	as_symbol = ['1', '2', '3', '4', '5', '6', '7', '8', '9', 'T', 'J', 'Q', 'K', 'A']

	from_symbol = {
		'1': ONE,
		'2': TWO,
		'3': THREE,
		'4': FOUR,
		'5': FIVE,
		'6': SIX,
		'7': SEVEN,
		'8': EIGHT,
		'9': NINE,
		'T': TEN,
		'J': JACK,
		'Q': QUEEN,
		'K': KING,
		'A': ACE
	}

class PokerMoves:
	""" Available poker moves
	
	Raise move may be changed
	at any time
	"""

	FOLD = 0
	CHECK = 1
	CALL = 2
	RAISE_ANY = 3
	RAISE_TEN = 3
	RAISE_QUARTER = 4
	RAISE_HALF = 5
	ALL_IN = 6

	NUM_MOVES = 7
	NUM_RAISE_MOVES = 3

	as_string = ['Fold', 'Check', 'Call', 'Raise 10%', 'Raise 25%', 'Raise half', 'All-in']

class PlayerState:
	""" Enum with the states of a poker player
	
	A player can be either playing or
	broken. A playing player can be:
	- folded, if he folded in this hand;
	- active, if he gets to bet in the
		current turn;
	- called, if he called or raised;
	- all-in, if he bet all he had.
	"""

	FOLDED = 0
	ACTIVE = 1
	CALLED = 2
	ALL_IN = 3
	BROKEN = 4

	NUM_STATES = 5

	as_string = ['Folded', 'Active', 'Called', 'All-in', 'Broken']


import numpy as np
from .agent import PokerAgent, Game

class RandomAgent(PokerAgent):
	""" Totally randomized agent

	Implements a poker agent that selects
	a random action from the list of valid
	actions using a uniform distribution
	"""

	def __call__(self, state: Game.StateView) -> int:
		"""  """
		
		# Return random valid action
		vu = state.valid_actions
		return np.random.choice(len(vu), p=vu / np.sum(vu))


from typing import List, Tuple
from itertools import islice
from functools import reduce
from .cards import Card
from .enums import HandRanking, CardRank, CardSuit

def eval_hand(hand: List[Card]) -> Tuple[int, List[int]]:
	""" Evaluate hand of cards

	Params
	------
	hand : list of cards
		Up to seven cards to evaluate
		as a poker hand.
	
	Returns
	-------
	tuple
		A tuple where the first item
		is the hand ranking, as in
		HandRanking enum, and the second
		is a list of kickers used to
		break ties. The number of kickers
		depends on the type of hand; for
		instance, only one kicker is
		returned for a straight, and five
		are returned for a flush.
	"""

	if not hand: return HandRanking.NONE, []
	elif len(hand) == 1: return HandRanking.HIGH, [hand[0].rank]
	elif len(hand) == 2:
		first, second = hand
		if first.rank == second.rank: return HandRanking.PAIR, [first.rank]
		else: return HandRanking.HIGH, [max(first.rank, second.rank), min(first.rank, second.rank)]
	else:
		# Sort by rank
		rank_sorted = sorted(hand, key=lambda c: c.rank, reverse=True)
		suit_sorted = sorted(hand, key=lambda c: (c.suit << 4) | c.rank, reverse=True)

		idx = 0
		flush = CardSuit.NUM_SUITS
		both = CardSuit.NUM_SUITS
		kind = 0
		straight = 0
		fourakind = []
		threeakind = []
		twoakind = []

		for rank, suit in zip(rank_sorted, suit_sorted):			
			# This block checks flush and straight flush
			if suit.suit == (flush & 0xf):
				flush += 0x100

				# Check straigth flush
				if suit.rank + (both >> 8) == (both >> 4 & 0xf): both += 0x100
				else: both = 0x100 | (suit.rank << 4) | suit.suit
			elif (flush >> 8) < 5: both = flush = 0x100 | (suit.rank << 4) | suit.suit; flush_start = idx
			
			# This block checks N of a kind and straights
			if rank.rank == (kind & 0xf): kind += 0x10
			else:
				# Save as pair, tris or poker
				numakind = kind >> 4
				if numakind == 2: twoakind.append(kind & 0xf)
				elif numakind == 3: threeakind.append(kind & 0xf)
				elif numakind == 4: fourakind.append(kind & 0xf)

				# Reset kind and check straight
				kind = 0x10 | rank.rank
				if rank.rank + (straight >> 4) == (straight & 0xf): straight += 0x10
				elif (straight >> 4) < 5: straight = 0x10 | rank.rank
			
			idx += 1
		
		# Check last kind
		numakind = kind >> 4
		if numakind == 2: twoakind.append(kind & 0xf)
		elif numakind == 3: threeakind.append(kind & 0xf)
		elif numakind == 4: fourakind.append(kind & 0xf)

		# Handle special case of 5-A straight
		if (both >> 8) == 4 and (both >> 4 & 0xf) == CardRank.FIVE:
			ace = next((True for c in hand if c.rank == CardRank.ACE and c.suit == (both & 0xf)), False)
			if ace: return HandRanking.STRAIGHT_FLUSH, [CardRank.FIVE]
		elif (straight >> 4) == 4 and (straight & 0xf) == CardRank.FIVE:
			ace = next((True for c in hand if c.rank == CardRank.ACE), False)
			if ace: return HandRanking.STRAIGHT, [CardRank.FIVE]
		
		if (both >> 8) >= 5: return HandRanking.STRAIGHT_FLUSH, [both >> 4 & 0xf]
		elif fourakind: return HandRanking.POKER, [fourakind[0], *islice((c.rank for c in rank_sorted if c.rank != fourakind[0]), 1)]
		elif len(threeakind) > 1: return HandRanking.FULL, [threeakind[0], threeakind[1]]
		elif threeakind and twoakind: return HandRanking.FULL, [threeakind[0], twoakind[0]]
		elif (flush >> 8) >= 5: return HandRanking.FLUSH, [c.rank for c in suit_sorted[flush_start:flush_start + 5]]
		elif (straight >> 4) >= 5: return HandRanking.STRAIGHT, [straight & 0xf]
		elif threeakind: return HandRanking.TRIS, [threeakind[0], *islice((c.rank for c in rank_sorted if c.rank != threeakind[0]), 2)]
		elif len(twoakind) > 1: return HandRanking.TWO_PAIR, [twoakind[0], twoakind[1], *islice((c.rank for c in rank_sorted if c.rank != twoakind[0] and c.rank != twoakind[1]), 1)]
		elif twoakind: return HandRanking.PAIR, [twoakind[0], *islice((c.rank for c in rank_sorted if c.rank != twoakind[0]), 3)]
		else: return HandRanking.HIGH, [c.rank for c in rank_sorted[:5]]

def get_kickers_value(kickers) -> int:
	""" Returns the kickers value as a bit-packed integer
	
	For instance, the kickers `[9, 8, 3]`
	are returned as `(9 << 8) | (8 << 4)
	| 3`
	"""

	return reduce(lambda value, kicker: value | (kicker[1] << (kicker[0] << 2)), enumerate(reversed(kickers)), 0)

def compare_rankings(rankings: List[Tuple[int, List[int]]]) -> Tuple[List[int], List[int], Tuple[int, List[int]]]:
	""" Compares multiple hands
	
	Returns the winner and the
	rankings of each hand

	Params
	------
	`rankings` : list of hand rankings
		A list of hand rankings with
		kickers, as output by `eval_hand`
	
	Returns
	-------
	tuple
		Returns a tuple where:
		- the first element is a one-hot
			encoded array of the winners;
		- the second element is a list
			with the indices of the winner.
	"""

	winners = []
	best_rank = HandRanking.NONE
	best_kicker = 0
	idx = 0

	for rank, kickers in rankings:
		kicker = get_kickers_value(kickers)
		if rank < best_rank:
			# Flush winners
			best_rank = rank
			best_kicker = kicker
			winners = [idx]
		elif rank == best_rank:
			if kicker > best_kicker:
				# Flush winners
				kicker = best_kicker
				winners = [idx]
			elif kicker == best_kicker:
				# Add winner
				winners.append(idx)
		
		# Next hand
		idx += 1

	onehot = [int(idx in winners) for idx, _ in enumerate(rankings)]
	return onehot, winners

def compare_hands(hands: List[List[Card]]) -> Tuple[List[int], List[int], Tuple[int, List[int]]]:
	""" Compares multiple hands
	
	Returns the winner and the
	rankings of each hand

	Params
	------
	`hand` : list of hands
		A list of lists of cards, one for
		each player to compare. Ideally they
		should all have the same number of
		cards
	
	Returns
	-------
	tuple
		Returns a tuple where:
		- the first element is a one-hot
			encoded array of the winners;
		- the second element is a list
			with the indices of the winner;
		- the last element is a list of
			hand rankings, as output by
			the `eval_hand` function, one
			for each hand.
	"""

	rankings = [eval_hand(hand) for hand in hands]
	return compare_rankings(rankings) + (rankings,)


from setuptools import setup, find_packages
from pathlib import Path

# The project root directory
root_dir = Path(__file__).parent

# Description and README
description = 'RL environments for poker games'
readme = (root_dir / 'README.md').read_text()

# Versioning
MAJOR = 0
MINOR = 0
PATCH = 6
version = '{:d}.{:d}.{:d}'.format(MAJOR, MINOR, PATCH)
pyversion = '>=3.6'

# Repo and author
author = 'Andrea Mecchia, Guglielmo Manneschi',
author_email = 'mecchia.andrea@gmail.com'
github = 'https://github.com/sneppy/pokerl'
lic = 'MIT'

# PyPi classifiers
classifiers = [
	'Development Status :: 1 - Planning',
	'Operating System :: OS Independent',
	'Programming Language :: Python :: 3.6',
	'Programming Language :: Python :: 3.7',
	'Programming Language :: Python :: 3.8',
	'Programming Language :: Python :: 3.9',
	'Topic :: Scientific/Engineering :: Artificial Intelligence',
	'Intended Audience :: Science/Research',
	'License :: OSI Approved :: MIT License',
	'Natural Language :: English'
]

# Packages and dependencies
packages = [p for p in find_packages() if p.startswith('pokerl')]
dependencies = ['numpy >= 1.18.0']

setup(
	name='pokerl',
	description=description,
	long_description=readme,
	long_description_content_type='text/markdown',
	version=version,
	python_requires=pyversion,
	author=author,
	author_email=author_email,
	url=github,
	license=lic,
	classifiers=classifiers,
	packages=packages,
	install_requires=dependencies,
	zip_safe=False
)


import socket, pickle
from ..game import Game

class PokerGameServer:
	"""  """

	def __init__(self, **game_config):
		"""  """

		self.game = Game(**game_config)
		self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.clients = []
	
	def run(self, host: str='localhost', port: int=25560):
		"""  """
		
		# Waiting for players
		self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		self.sock.bind((host, int(port)))
		self.sock.listen()

		print('server running on %s:%d' % (host, int(port)))

		while len(self.clients) < self.game.num_players:
			client, addr = self.sock.accept()
			print('%s connecting as Player %d' % (addr, len(self.clients)))
			self.clients.append(client)
		
		# Start game
		self.game.reset()
		for player, client in enumerate(self.clients):
			# Send ready state
			data = pickle.dumps(dict(status='ready', player=player, num_players=self.game.num_players))
			client.sendall(data)
		
		game_over = False
		while not game_over:
			player_client = self.clients[self.game.active_player]

			# Get active state and send to player
			active_state = self.game.active_state
			active_state = pickle.dumps(dict(status='playing', player=self.game.active_player, active_state=active_state))
			player_client.sendall(active_state)

			# Wait for a response
			action = player_client.recv(1024)
			player, action = pickle.loads(action)
			assert player == self.game.active_player, 'Received action from wrong player'
			
			# Do action
			game_over, _, _ = self.game.step(action)
		
		print('game ended')